from enum import StrEnum, IntEnum


class PaymentType(StrEnum):
    DEBIT = "debit"
    HOLD = "hold"


class PaymentScheme(StrEnum):
    FULL = "full"
    BNPL_PARTS_4 = "bnpl_parts_4"
    BNPL_LATER_30 = "bnpl_later_30"


class PaymentMethod(StrEnum):
    PAN = "pan"
    APPLE = "apple"
    GOOGLE = "google"
    MONOBANK = "monobank"


class AmountType(StrEnum):
    MERCHANT = "merchant"
    CLIENT = "client"
    FIX = "fix"


class TransactionStatus(StrEnum):
    CREATED = "created"
    PROCESSING = "processing"
    HOLD = "hold"
    SUCCESS = "success"
    FAILURE = "failure"
    REVERSER = "reversed"
    EXPIRED = "expired"


class CancellationStatus(StrEnum):
    PROCESSING = "processing"
    SUCCESS = "success"
    FAILURE = "failure"


class TokenizedCardStatus(StrEnum):
    NEW = "new"
    CREATED = "created"
    FAILED = "failed"


class AccountType(StrEnum):
    BLACK = "black"
    WHITE = "white"
    PLATINUM = "platinum"
    IRON = "iron"
    FOP = "fop"
    YELLOW = "yellow"
    eAID = "eAid"


class CashbackType(StrEnum):
    NONE = "None"
    UAH = "UAH"
    MILES = "Miles"


class RegistrationStatus(StrEnum):
    NEW = "New"
    DECLINED = "Declined"
    APPROVED = "Approved"


class CurrencyCode(IntEnum):
    UAH = 980
    USD = 840
    EUR = 978
    GBP = 826
    JPY = 392
    CHF = 756
    CNY = 156
    AED = 784
    AFN = 971
    ALL = 8
    AMD = 51
    AOA = 973
    ARS = 32
    AUD = 36
    AZN = 944
    BDT = 50
    BGN = 975
    BHD = 48
    BIF = 108
    BND = 96
    BOB = 68
    BRL = 986
    BWP = 72
    BYN = 933
    CAD = 124
    CDF = 976
    CLP = 152
    COP = 170
    CRC = 188
    CUP = 192
    CZK = 203
    DJF = 262
    DKK = 208
    DZD = 12
    EGP = 818
    ETB = 230
    GEL = 981
    GHS = 936
    GMD = 270
    GNF = 324
    HKD = 344
    HRK = 191
    HUF = 348
    IDR = 360
    ILS = 376
    INR = 356
    IQD = 368
    IRR = 364
    ISK = 352
    JOD = 400
    KES = 404
    KGS = 417
    KHR = 116
    KPW = 408
    KRW = 410
    KWD = 414
    KZT = 398
    LAK = 418
    LBP = 422
    LKR = 144
    LYD = 434
    MAD = 504
    MDL = 498
    MGA = 969
    MKD = 807
    MNT = 496
    MRO = 478
    MUR = 480
    MVR = 462
    MWK = 454
    MXN = 484
    MYR = 458
    MZN = 943
    NAD = 516
    NGN = 566
    NIO = 558
    NOK = 578
    NPR = 524
    NZD = 554
    OMR = 512
    PEN = 604
    PHP = 608
    PKR = 586
    PLN = 985
    PYG = 600
    QAR = 634
    RON = 946
    RSD = 941
    SAR = 682
    SCR = 690
    SDG = 938
    SEK = 752
    SGD = 702
    SLL = 694
    SOS = 706
    SRD = 968
    SYP = 760
    SZL = 748
    THB = 764
    TJS = 972
    TMT = 795
    TND = 788
    TRY = 949
    TWD = 901
    TZS = 834
    UGZ = 800
    UYU = 858
    UZS = 860
    VEF = 937
    VND = 704
    XAF = 950
    XDR = 960
    XOF = 952
    YER = 886
    ZAR = 710
    ZMK = 894



__all__ = [
    'create_privkey_pubkey'
]

import ecdsa
import hashlib
import codecs


def create_privkey_pubkey(privkey_path: str, pubkey_path: str) -> None:
    private_key = ecdsa.SigningKey.generate(curve=ecdsa.SECP256k1, hashfunc=hashlib.sha256)
    public_key = private_key.get_verifying_key()

    private_key_pem = private_key.to_pem()
    public_key_pem = public_key.to_pem()

    with codecs.open(privkey_path, 'w', 'utf-8') as f:
        f.write(private_key_pem.decode('utf-8'))

    with codecs.open(pubkey_path, 'w', 'utf-8') as f:
        f.write(public_key_pem.decode('utf-8'))



from abc import ABC, abstractmethod

__all__ = [
    'APIMethod'
]

import base64
import random
import string

import ecdsa
import hashlib

import re
import json
import requests

from enum import Enum, StrEnum
from datetime import datetime

from pydantic import BaseModel

from requests import Response
from aiohttp import (
    ClientSession,
    ClientResponse
)
from typing import Union, Optional

from mbnk.exceptions import *

from mbnk.types import *

from mbnk.responses import *

from pydantic import dataclasses


class BaseAPIMethod(ABC):
    _api_token = None
    _headers = {}

    def __init__(
            self,
            base_url: str,
            api_token: Optional[str] = None
    ):
        self._base_url: str = base_url

        if api_token is not None:
            self._api_token: str = api_token
            self._headers["X-Token"] = self._api_token

    @staticmethod
    def _time_header():
        def outer(func):
            def inner(self):
                self.__headers["X-Time"]: str = str(int(datetime.now().timestamp()))

            return inner

        return outer

    @staticmethod
    def _key_id_header():
        def outer(func):
            def inner(self):
                self.__headers["X-Key-Id"]: str = self.__key_id

            return inner

        return outer

    @abstractmethod
    def _request(self):
        pass


class APIMethod(BaseAPIMethod):

    def __init__(
            self,
            base_url: str,
            api_token: Optional[str] = None,
    ):
        super().__init__(
            base_url=base_url,
            api_token=api_token
        )

    @staticmethod
    def __is_exception(response: Union[Response, ClientResponse]) -> bool:
        if isinstance(response, Response):
            status_code = response.status_code
        elif isinstance(response, ClientResponse):
            status_code = response.status
        else:
            return True

        if status_code != 200:
            return True

        return False

    def __sync_request(
            self,
            method: str,
            path: str,
            data: str = None,
            params: str = None
    ):
        response = requests.request(
            method=method,
            url=f"{self.__base_url}/{path}",
            headers=self.__headers,
            params=params,
            data=json.dumps(data) if data is not None else None
        )
        response_data = response.json()

        if self.__is_exception(response):
            raise MonobankAPIException

        if isinstance(response_data, list):
            return {"list": response_data}

        return response.json()

    async def __async_request(
            self,
            method: str,
            path: str,
            data: str = None,
            params: str = None
    ):
        async with ClientSession() as session:
            request = getattr(session, method)
            async with request(
                    url=f"{self.__base_url}/{path}",
                    headers=self.__headers,
                    data=json.dumps(data) if data is not None else None,
                    params=params
            ) as response:
                response_data = await response.json()
                response_data = self.__load_response(response_data)

                if self.__is_exception(response):
                    raise MonobankAPIException

                return response_data

    @staticmethod
    def __generate_request_id():
        request_id = "".join(random.choices(string.ascii_letters + string.digits, k=random.randint(18, 24)))
        return request_id

    @staticmethod
    def _request_id_header():
        def outer(func):
            def inner(self):
                self.__headers["X-Request-Id"]: str = self.__generate_request_id()

            return inner

        return outer

    @staticmethod
    def __create_signature(self):
        url = "https://api.monobank.ua"
        data = (self.__headers["X-Time"] + url).encode('utf-8')

        private_key = ecdsa.SigningKey.from_pem(self.__privkey, hashfunc=hashlib.sha256)

        sign = private_key.sign(data, hashfunc=hashlib.sha256)
        sign_base64 = base64.b64encode(sign)

        return sign_base64

    @staticmethod
    def _sign_header():
        def outer(func):
            def inner(self):
                self.__headers["X-Sign"]: str = self.__create_signature(self)
            return inner

        return outer

    @staticmethod
    async def _async_request(
            method: str,
            path: str
    ):
        async def outer(func):
            async def inner(*args, **kwargs):
                self = args[0]
                args = args[1:]

                func_args = {
                    "method": method,
                    "path": path,
                    ("params" if request_method == "get" else "data"): self.__build_data(**kwargs)
                }

                async with ClientSession() as session:
                    async with session.request(
                            method=method,
                            url=f"{self.__base_url}/{path}",
                            headers=self.__headers,
                            data=json.dumps(data) if data is not None else None,
                            params=params
                    ) as response:
                        response_data = await response.json()
                        response_data = self.__load_response(response_data)

                        if self.__is_exception(response):
                            raise MonobankAPIException

                        if isinstance(response, MonobankAPIException) or isinstance(response, MonobankAPIException):
                            return response

                        return func(self, *args, **kwargs, response_data=response)

            return inner

        return outer

    @staticmethod
    def _request(
            method: str,
            path: str
    ):
        def outer(func):
            def inner(*args, **kwargs):
                self = args[0]
                args = args[1:]

                func_args = {
                    "method": method,
                    "path": path,
                    ("params" if request_method == "get" else "data"): self.__build_data(**kwargs)
                }

                response = requests.request(
                    method=method,
                    url=f"{self.__base_url}/{path}",
                    headers=self.__headers,
                    params=params,
                    data=json.dumps(data) if data is not None else None
                )

                response_data = response.json()

                if self.__is_exception(response):
                    raise MonobankAPIException

                if isinstance(response_data, list):
                    response_data = {"list": response_data}

                return func(self, *args, **kwargs, response_data=response_data)

            return inner

        return outer



__all__ = [
    # Monobank Open API Types
    'Currency',
    'Account',
    'Jar',
    'Transaction',
    'CurrencyList',
    'ClientInfo',
    'Statement',

    # Monobank Corporate Open API Types

    # Mono Acquiring API Types
    'QrListItem',
    'QrList',
    'QrDetails',
    'InvoiceCreated',
    'InvoiceCanceled',
    'InvoiceStatus',
    'InvoiceInfo',
    'FinalizeInvoice',
    'Product',
    'MerchantPaymInfo',
    'SaveCardData',
    'CancelListItem',
    'WalletItem',
    'WalletCards',
    'MerchantStatementItem',
    'MerchantDetails',
    'MerchantStatement',
    'MerchantPubKey'
]

from typing import List, Optional, Union
from pydantic import BaseModel, Field
from pydantic.dataclasses import dataclass
from mbnk.enums import *


# Monobank Open API Types

# Public
@dataclass
class Currency(BaseModel):
    currency_code_a: CurrencyCode = Field(alias='currencyCodeA')
    currency_code_b: CurrencyCode = Field(alias='currencyCodeB')
    date: int
    rate_sell: float = Field(alias='rateSell')
    rate_buy: float = Field(alias='rateBuy')
    rate_cross: float = Field(alias='rateCross')


@dataclass
class CurrencyList(BaseModel):
    list: List[Currency]


# Personal
@dataclass
class Account(BaseModel):
    id: str
    send_id: str = Field(alias="sendId")
    balance: int
    credit_limit: int = Field(alias="creditLimit")
    type: AccountType
    currency_code: CurrencyCode = Field(alias="currencyCode")
    masked_pan: List[str] = Field(alias="maskedPan")
    iban: str
    cashback_type: Optional[CashbackType] = Field(default=None, alias="cashbackType")


@dataclass
class Jar(BaseModel):
    id: str
    send_id: str = Field(alias="sendId")
    title: str
    description: str
    currency_code: CurrencyCode = Field(alias="currencyCode")
    balance: int
    goal: int


@dataclass
class Transaction(BaseModel):
    id: str
    time: int
    description: str
    mcc: int
    original_mcc: int = Field(alias="originMcc")
    hold: bool
    amount: int
    operation_amount: int = Field(alias="operationAmount")
    currency_code: CurrencyCode = Field(alias="currencyCode")
    commission_rate: int = Field(alias="commissionRate")
    cashback_amount: int = Field(alias="cashbackAmount")
    balance: int
    comment: Optional[str] = None
    receipt_id: Optional[str] = Field(default=None, alias="receiptId")
    invoice_id: Optional[str] = Field(default=None, alias="invoiceId")
    counter_edrpou: Optional[str] = Field(default=None, alias="counterEdrpou")
    counter_iban: Optional[str] = Field(default=None, alias="counterIban")
    counter_name: Optional[str] = Field(default=None, alias="counterName")


class ClientInfo(BaseModel):
    client_id: str = Field(alias="clientId")
    name: str
    web_hook_url: str = Field(alias="webHookUrl")
    permissions: str
    accounts: List[Account]
    jars: List[Jar]


class Statement(BaseModel):
    list: List[Transaction]


# Mono Acquiring API

# Qr
class QrListItem(BaseModel):
    short_qr_id: str = Field(alias="shortQrId")
    qr_id: str = Field(alias="qrId")
    amount_type: AmountType = Field(alias="amountType")
    page_url: str = Field(alias="pageUrl")


class QrList(BaseModel):
    list: List[QrListItem]


class QrDetails(BaseModel):
    short_qr_id: str = Field(alias="shortQrId")
    invoice_id: Optional[str] = Field(default=None, alias="invoiceId")
    amount: Optional[int] = Field(default=None)
    ccy: Optional[int] = Field(default=None)


# Invoice
class CanceledItem(BaseModel):
    status: CancellationStatus
    created_date: str = Field(alias="createdDate")
    modified_date: str = Field(alias="modifiedDate")
    amount: Optional[int] = Field(default=None)
    ccy: Optional[CurrencyCode] = Field(default=None)
    approval_code: Optional[str] = Field(default=None, alias="approvalCode")
    rrn: Optional[str] = Field(default=None)
    ext_ref: Optional[str] = Field(default=None, alias="extRef")


class InvoiceCreated(BaseModel):
    invoice_id: str = Field(alias="invoiceId")
    page_url: str = Field(alias="pageUrl")


class InvoiceCanceled(BaseModel):
    status: str
    created_date: str = Field(alias="createdDate")
    modified_date: str = Field(alias="modifiedDate")


class WalletData(BaseModel):
    card_token: str = Field(alias="cardToken")
    wallet_id: str = Field(alias="walletId")
    status: TokenizedCardStatus


class InvoiceStatus(BaseModel):
    invoice_id: str
    status: TransactionStatus
    amount: int
    ccy: CurrencyCode
    failure_reason: Optional[str] = None
    final_amount: Optional[int] = None
    created_date: Optional[str] = None
    modified_date: Optional[str] = None
    reference: Optional[str] = None
    cancel_list: Optional[CanceledItem] = None
    wallet_data: Optional[WalletData] = None


class InvoiceInfo(BaseModel):
    masked_pan: str
    approval_code: str
    rrn: str
    amount: int
    ccy: CurrencyCode
    final_amount: int
    terminal: str
    payment_scheme: PaymentScheme
    payment_method: PaymentMethod
    domestic_card: bool
    country: str
    # created_date = Optional[str] = None
    fee: Optional[int] = None
    cancel_list: Optional[CanceledItem] = None


@dataclass
class FinalizeInvoice:
    status: str


class Product(BaseModel):
    name: str
    qty: float
    sum: int
    icon: Optional[str] = None
    unit: Optional[str] = None
    code: Optional[str] = None
    barcode: Optional[str] = None
    header: Optional[str] = None
    footer: Optional[str] = None
    tax: Optional[List[int]] = None
    uktzed: Optional[str] = None


class MerchantPaymInfo(BaseModel):
    reference: str
    destination: str
    basket_order: List[Product] = Field(alias="basketOrder")


@dataclass
class SaveCardData(BaseModel):
    save_card: bool = Field(alias="saveCard")
    wallet_id: Optional[str] = Field(default=None, alias="walletId")


@dataclass
class CancelListItem(BaseModel):
    amount: int
    ccy: CurrencyCode
    date: str
    masked_pan: str = Field(default=None, alias="maskedPan")
    approval_code: Optional[str] = Field(default=None, alias="approvalCode")
    rrn: Optional[str] = Field(default=None)


# Merchant
@dataclass
class MerchantStatementItem(BaseModel):
    invoice_id: str = Field(alias="invoiceId")
    status: str
    masked_pan: str = Field(alias="maskedPan")
    date: str
    payment_scheme: PaymentScheme = Field(alias="paymentScheme")
    amount: int
    ccy: CurrencyCode
    profit: Optional[int] = Field(default=None)
    approval_code: Optional[str] = Field(default=None, alias="approvalCode")
    reference: Optional[str] = Field(default=None)
    cancel_list: Optional[List[CancelListItem]] = Field(default=None, alias="cancelList")
    rrn: Optional[str] = Field(default=None)
    short_qr_id: Optional[str] = Field(default=None, alias="shortQrId")


@dataclass
class MerchantDetails(BaseModel):
    merchant_id: str = Field(alias="merchantId")
    merchant_name: str = Field(alias="merchantName")


@dataclass
class MerchantStatement(BaseModel):
    list: List[MerchantStatementItem]


@dataclass
class MerchantPubKey(BaseModel):
    key: str


# Wallet
@dataclass
class WalletItem(BaseModel):
    card_token: str = Field(alias="cardToken")
    masked_pan: str = Field(alias="maskedPan")
    country: Optional[str] = Field(default=None)


@dataclass
class WalletCards(BaseModel):
    wallets: List[WalletItem]



__all__ = [
    'MonoAcquiringAPI',
    'MonobankOpenAPI',
    'MonobankCorporateOpenAPI'
]

from typing import Optional

from mbnk.api import (
    APIMethod,
    APIPaths
)


# Monobank Acquiring API


class Merchant(APIMethod):

    @APIMethod._request("get", path=APIPaths.merchant_details)
    def details(self, **kwargs) -> Union[MerchantDetails, MonobankAPIException]:
        """
        Mono Acquiring API: https://api.monobank.ua/docs/acquiring.html#/paths/~1api~1merchant~1details/get

        :return: MerchantDetailsResponse
        """

        return MerchantDetails(**kwargs["response_data"])

    @APIMethod._request("get", path=APIPaths.merchant_statement)
    def statement(
            self,
            from_timestamp: int,
            to_timestamp: Optional[int] = None,
            **kwargs
    ) -> Union[MerchantStatement, MonobankAPIException]:
        """
        Mono Acquiring API: https://api.monobank.ua/docs/acquiring.html#/paths/~1api~1merchant~1statement/get

        :param from_timestamp:
        :param to_timestamp:
        :param kwargs:
        :return:
        """

        return MerchantStatement(
            list=[
                MerchantStatementItem(
                    **statement_item,
                    cancel_list=[
                        CancelListItem(
                            **cancel_item
                        ) for cancel_item in statement_item['cancel_list']
                    ] if "cancel_list" in statement_item else None
                ) for statement_item in kwargs['response_data']['list']
            ]
        )

    @APIMethod._request("get", path=APIPaths.merchant_pubkey)
    def pubkey(self, **kwargs) -> Union[MerchantPubKey, MonobankAPIException]:
        """
        Mono Acquiring API Docs: https://api.monobank.ua/docs/acquiring.html#/paths/~1api~1merchant~1pubkey/get

        :return: MerchantPubKeyResponse
        """

        return MerchantPubKey(**kwargs["response_data"])


class Invoice(APIMethod):

    @APIMethod._request("post", path=APIPaths.invoice_create)
    def create(
            self,
            amount: int,
            ccy: Optional[int] = None,
            merchant_paym_info: Optional[MerchantPaymInfo] = None,
            redirect_url: Optional[str] = None,
            web_hook_url: Optional[str] = None,
            validity: Optional[int] = None,
            payment_type: Optional[str] = None,
            qr_id: Optional = None,
            save_card_data: Optional[SaveCardData] = None,
            **kwargs
    ) -> Union[InvoiceCreated, MonobankAPIException]:
        """
        Mono Acquiring API: https://api.monobank.ua/docs/acquiring.html#/paths/~1api~1merchant~1invoice~1create/post

        :param amount:
        :param ccy:
        :param merchant_paym_info:
        :param redirect_url:
        :param web_hook_url:
        :param validity:
        :param payment_type:
        :param qr_id:
        :param save_card_data:
        :return:
        """

        return InvoiceCreated(**kwargs['response_data'])

    @APIMethod._request("post", path=APIPaths.invoice_cancel)
    def cancel(
            self,
            invoice_id: str,
            ext_ref: str = None,
            amount: int = None,
            items=None,
            **kwargs
    ) -> Union[InvoiceCanceled, MonobankAPIException]:
        """
        Mono Acquiring API: https://api.monobank.ua/docs/acquiring.html#/paths/~1api~1merchant~1invoice~1cancel/post

        :param invoice_id:
        :param ext_ref:
        :param amount:
        :param items:
        :param kwargs:
        :return:
        """
        return InvoiceCanceled(**kwargs['response_data'])

    def status(
            self,
            invoice_id: str,
            **kwargs
    ) -> Union[InvoiceStatus, MonobankAPIException]:
        """
        Mono Acquiring API: https://api.monobank.ua/docs/acquiring.html#/paths/~1api~1merchant~1invoice~1status?invoiceId=%7BinvoiceId%7D/get

        :param invoice_id:
        :param kwargs:
        :return:
        """

        return kwargs["response_data"]

    @APIMethod._request("post", path=APIPaths.invoice_invalidation)
    def invalidation(
            self,
            invoice_id: str,
            **kwargs
    ) -> Union[EmptyResponse, MonobankAPIException]:
        """
        Source: https://api.monobank.ua/docs/acquiring.html#/paths/~1api~1merchant~1invoice~1remove/post

        :param invoice_id:
        :param kwargs:
        :return:
        """

        return EmptyResponse()

    @APIMethod._request("get", path=APIPaths.invoice_info)
    async def info(
            self,
            invoice_id: str,
            **kwargs
    ) -> Union[InvoiceInfo, MonobankAPIException]:
        """
        Source: https://api.monobank.ua/docs/acquiring.html#/paths/~1api~1merchant~1invoice~1payment-info?invoiceId=%7BinvoiceId%7D/get

        :param invoice_id:
        :param kwargs:
        :return:
        """

        return kwargs['response_data']

    @APIMethod._request("get", path=APIPaths.invoice_finalize)
    def finalize(
            self,
            invoice_id: str,
            amount: int,
            **kwargs
    ) -> Union[FinalizeInvoice, MonobankAPIException]:
        """
        Source: https://api.monobank.ua/docs/acquiring.html#/paths/~1api~1merchant~1invoice~1finalize/post

        :param invoice_id:
        :param amount:
        :param kwargs:
        :return:
        """
        return FinalizeInvoice(
            **kwargs['response_data']
        )


class Qr(APIMethod):

    @APIMethod._request("get", path=APIPaths.qr_list)
    def list(self, **kwargs) -> Union[QrList, MonobankAPIException]:
        return QrList(
            list=[
                QrListItem(
                    **qr
                ) for qr in kwargs['response']['list']
            ]
        )

    @APIMethod._request("post", path=APIPaths.qr_details)
    def details(
            self,
            qr_id: str,
            **kwargs
    ) -> Union[QrDetails, MonobankAPIException]:
        """
        :param qr_id:
        :return:
        """
        return QrDetails(**kwargs['response_data'])

    @APIMethod._request("post", path=APIPaths.qr_reset_amount)
    def reset_amount(
            self,
            qr_id: str,
            **kwargs
    ) -> Union[EmptyResponse, MonobankAPIException]:
        """
        :param qr_id:
        :return:
        """
        return EmptyResponse()


class Wallet(APIMethod):

    @APIMethod._request("get", path=APIPaths.wallet_cards)
    async def cards(
            self,
            wallet_id: str,
            **kwargs
    ):
        """
        Source: https://api.monobank.ua/docs/acquiring.html#/paths/~1api~1merchant~1wallet/get

        :param wallet_id:
        :param kwargs:
        :return:
        """
        return kwargs["response_data"]

    @APIMethod._request("post", path=APIPaths.wallet_payment)
    async def payment(self):
        """
        Source: https://api.monobank.ua/docs/acquiring.html#/paths/~1api~1merchant~1wallet~1payment/post
        :return:
        """
        pass

    @APIMethod._request("delete", path=APIPaths.wallet_delete_card)
    def delete_card(
            self,
            card_token: str,
            **kwargs
    ):
        """
        This method delete tokenized card from wallet
        Source: https://api.monobank.ua/docs/acquiring.html#/paths/~1api~1merchant~1wallet~1card/delete

        :param card_token:
        :param kwargs:
        :return:
        """
        return kwargs["response_data"]


class MonoAcquiringAPI(MonoAcquiringAPIModel):
    """
    Synchronous version MonoAcquiringAPI
    Source: https://api.monobank.ua/docs/acquiring.html
    """

    __base_url: str = "https://api.monobank.ua/api"

    def __init__(self, api_token: str):

        self.__api_token: str = api_token

        kwargs = {
            'api_token': self.__api_token,
            'base_url': self.__base_url
        }

        self.merchant: Merchant = Merchant(**kwargs)

        self.invoice: Invoice = Invoice(**kwargs)

        self.qr: Qr = Qr(**kwargs)

        self.wallet: Wallet = Wallet(**kwargs)

    def get_api_token(self):
        return self.__api_token


# Monobank Open API


class Public(APIMethod):

    @APIMethod._request("get", path=APIPaths.currencies_list)
    def currency(self, **kwargs) -> Union[CurrencyList, MonobankAPIException]:

        return CurrencyList.model_validate(kwargs['response_data'])


class Personal(APIMethod):

    @APIMethod._request("get", path=APIPaths.personal_info)
    def info(self, **kwargs) -> Union[ClientInfo, MonobankAPIException]:
        """
        Source: https://api.monobank.ua/docs/#tag/Kliyentski-personalni-dani/paths/~1personal~1client-info/get

        :return:
        """

        return ClientInfo.model_validate(kwargs['response_data'])

    @APIMethod._request("post", path=APIPaths.personal_webhook)
    def set_web_hook(self, web_hook_url: str) -> Union[EmptyResponse, MonobankAPIException]:
        """
        Source: https://api.monobank.ua/docs/#tag/Kliyentski-personalni-dani/paths/~1personal~1webhook/post

        :param web_hook_url:
        :return:
        """
        return EmptyResponse()

    @APIMethod._request("get", path=APIPaths.personal_statement)
    def statement(self, **kwargs) -> Union[Statement, MonobankAPIException]:

        return Statement(
            list=[
                Transaction(
                    **transaction
                ) for transaction in kwargs['response_data']
            ]
        )


class MonobankOpenAPI:
    """
    Synchronous version MonobankOpenAPI
    Source: https://api.monobank.ua/docs/
    """

    __base_url = "https://api.monobank.ua"
    __api_token: str = None

    def __init__(self, api_token: Optional[str] = None):
        self.__api_token = api_token

        kwargs = {
            'base_url': self.__base_url,
            'api_token': self.__api_token
        }

        self.public: Public = Public(**kwargs)

        self.personal: Personal = Personal(**kwargs)

    def set_api_token(self, api_token: str):
        self.__api_token: str = api_token

    def get_api_token(self) -> str:
        return self.__api_token


# Monobank Corporate Open API


class Client(APIMethod):

    @APIMethod._request("post", path=APIPaths.init_access)
    def init_access(self):
        pass

    @APIMethod._request("get", path=APIPaths.check_access)
    def check_access(self):
        pass


class Authorization(APIMethod):

    @APIMethod._request("post", path=APIPaths.auth_registration)
    def registration(self):
        pass

    @APIMethod._request("post", path=APIPaths.auth_status)
    def status(self):
        pass

class MonobankCorporateOpenAPI:
    """
    Synchronous version MonobankCorporateOpenAPI
    Source: https://api.monobank.ua/docs/corporate.html
    """

    __base_url: str = "https://api.monobank.ua"
    __headers: dict = {}

    __key_id: str = None

    __priv_key: str = None
    __pub_key: str = None

    def __init__(self):

        self.public: Public = Public(
            base_url=self.__base_url
        )

        self.auth: Authorization = Authorization(
            base_url=self.__base_url
        )

        self.client: Client = Client(
            base_url=self.__base_url
        )

    def set_key_id(self, key_id: str):
        self.__key_id: str = key_id

    def set_priv_key(self, priv_key_path: str):
        with codecs.open(priv_key_path, 'r', 'utf-8') as f:
            priv_key = f.read()
            f.close()

        self.__priv_key = priv_key

    def set_pub_key(self, pub_key_path: str):
        with codecs.open(pub_key_path, 'r', 'utf-8') as f:
            pub_key = f.read()
            f.close()

        self.__pub_key = pub_key



__all__ = [
    'AsyncMonoAcquiringAPI',
    'AsyncMonobankOpenAPI',
    'AsyncMonobankCorporateOpenAPI'
]

from typing import Optional, Union

from mbnk.exceptions import MonobankAPIException
from mbnk.types import *
from mbnk.api import (
    APIMethod,
    APIPaths
)
from mbnk.responses import *


class Merchant(APIMethod):

    @APIMethod._request("get", path="/details")
    def details(self, **kwargs) -> Union[MerchantDetails, MonobankAPIException]:
        """
        Mono Acquiring API: https://api.monobank.ua/docs/acquiring.html#/paths/~1api~1merchant~1details/get

        :return: MerchantDetailsResponse
        """

        return MerchantDetails(**kwargs["response_data"])

    @APIMethod._request("get", path="/statement")
    def statement(
            self,
            from_timestamp: int,
            to_timestamp: Optional[int] = None,
            **kwargs
    ) -> Union[MerchantStatement, MonobankAPIException]:
        """
        Mono Acquiring API: https://api.monobank.ua/docs/acquiring.html#/paths/~1api~1merchant~1statement/get

        :param from_timestamp:
        :param to_timestamp:
        :param kwargs:
        :return:
        """

        return MerchantStatement(
            list=[
                MerchantStatementItem(
                    **statement_item,
                    cancel_list=[
                        CancelListItem(
                            **cancel_item
                        ) for cancel_item in statement_item['cancel_list']
                    ] if "cancel_list" in statement_item else None
                ) for statement_item in kwargs['response_data']['list']
            ]
        )

    @APIMethod._request("get", path="/pubkey")
    def pubkey(self, **kwargs) -> Union[MerchantPubKey, MonobankAPIException]:
        """
        Mono Acquiring API Docs: https://api.monobank.ua/docs/acquiring.html#/paths/~1api~1merchant~1pubkey/get

        :return: MerchantPubKeyResponse
        """

        return MerchantPubKey(**kwargs["response_data"])


class Invoice(APIMethod):

    @APIMethod._request("post", path="/create")
    async def create(
            self,
            amount: int,
            ccy: Optional[int] = None,
            merchant_paym_info: Optional[MerchantPaymInfo] = None,
            redirect_url: Optional[str] = None,
            web_hook_url: Optional[str] = None,
            validity: Optional[int] = None,
            payment_type: Optional[str] = None,
            qr_id: Optional = None,
            save_card_data: Optional[SaveCardData] = None,
            **kwargs
    ) -> Union[InvoiceCreated, MonobankAPIException]:
        """
        Mono Acquiring API: https://api.monobank.ua/docs/acquiring.html#/paths/~1api~1merchant~1invoice~1create/post

        :param amount:
        :param ccy:
        :param merchant_paym_info:
        :param redirect_url:
        :param web_hook_url:
        :param validity:
        :param payment_type:
        :param qr_id:
        :param save_card_data:
        :return:
        """

        return InvoiceCreated(**kwargs['response_data'])

    @APIMethod._request("post", path="/cancel")
    async def cancel(
            self,
            invoice_id: str,
            ext_ref: str = None,
            amount: int = None,
            items=None,
            **kwargs
    ) -> Union[InvoiceCanceled, MonobankAPIException]:
        """
        Mono Acquiring API: https://api.monobank.ua/docs/acquiring.html#/paths/~1api~1merchant~1invoice~1cancel/post

        :param invoice_id:
        :param ext_ref:
        :param amount:
        :param items:
        :param kwargs:
        :return:
        """
        return InvoiceCanceled(**kwargs['response_data'])

    @APIMethod._request("get", path="/status")
    async def status(
            self,
            invoice_id: str,
            **kwargs
    ) -> Union[InvoiceStatus, MonobankAPIException]:
        """
        Mono Acquiring API: https://api.monobank.ua/docs/acquiring.html#/paths/~1api~1merchant~1invoice~1status?invoiceId=%7BinvoiceId%7D/get

        :param invoice_id:
        :param kwargs:
        :return:
        """

        return kwargs["response_data"]

    @APIMethod._request("post", path="/remove")
    async def invalidation(
            self,
            invoice_id: str,
            **kwargs
    ) -> Union[EmptyResponse, MonobankAPIException]:
        """
        Source: https://api.monobank.ua/docs/acquiring.html#/paths/~1api~1merchant~1invoice~1remove/post

        :param invoice_id:
        :param kwargs:
        :return:
        """

        return EmptyResponse()

    @APIMethod._request("get", path="/info")
    async def info(
            self,
            invoice_id: str,
            **kwargs
    ) -> Union[InvoiceInfo, MonobankAPIException]:
        """
        Source: https://api.monobank.ua/docs/acquiring.html#/paths/~1api~1merchant~1invoice~1payment-info?invoiceId=%7BinvoiceId%7D/get

        :param invoice_id:
        :param kwargs:
        :return:
        """

        return kwargs['response_data']

    @APIMethod._async_request("get", path="/finalize")
    async def finalize(
            self,
            invoice_id: str,
            amount: int,
            **kwargs
    ) -> Union[FinalizeInvoice, MonobankAPIException]:
        """
        Source: https://api.monobank.ua/docs/acquiring.html#/paths/~1api~1merchant~1invoice~1finalize/post

        :param invoice_id:
        :param amount:
        :param kwargs:
        :return:
        """

        return FinalizeInvoice(
            **kwargs['response_data']
        )


class Qr(APIMethod):

    @APIMethod._request("get", path="/list")
    async def list(self, **kwargs) -> Union[QrList, MonobankAPIException]:
        return QrList(
            list=[
                QrListItem(
                    **qr
                ) for qr in kwargs['response']['list']
            ]
        )

    @APIMethod._request("get", path="/details")
    async def details(
            self,
            qr_id: str,
            **kwargs
    ) -> Union[QrDetails, MonobankAPIException]:
        """
        :param qr_id:
        :return:
        """
        return QrDetails(**kwargs['response_data'])

    @APIMethod._request("post", path="/reset-amount")
    async def reset_amount(
            self,
            qr_id: str,
            **kwargs
    ) -> Union[EmptyResponse, MonobankAPIException]:
        """
        :param qr_id:
        :return:
        """
        return EmptyResponse()


class Wallet(APIMethod):

    @APIMethod._request("get", path="")
    async def cards(
            self,
            wallet_id: str,
            **kwargs
    ):
        """
        Source: https://api.monobank.ua/docs/acquiring.html#/paths/~1api~1merchant~1wallet/get

        :param wallet_id:
        :param kwargs:
        :return:
        """

        return kwargs["response_data"]

    @APIMethod._request("post", path="/payment")
    async def payment(self):
        """
        Source: https://api.monobank.ua/docs/acquiring.html#/paths/~1api~1merchant~1wallet~1payment/post
        :return:
        """

        pass

    @APIMethod._request("delete", path="/card")
    async def delete_card(
            self,
            card_token: str,
            **kwargs
    ):
        """
        This method delete tokenized card from wallet
        Source: https://api.monobank.ua/docs/acquiring.html#/paths/~1api~1merchant~1wallet~1card/delete

        :param card_token:
        :param kwargs:
        :return:
        """

        return kwargs["response_data"]


class AsyncMonoAcquiringAPI:
    """
    Asynchronous version MonoAcquiringAPI
    Source: https://api.monobank.ua/docs/acquiring.html
    """

    __base_url: str = "https://api.monobank.ua/api"

    def __init__(self, api_token: str):
        self.__api_token: str = api_token

        kwargs = {
            "api_token": self.__api_token,
            "base_url": self.__base_url
        }

        self.merchant: Merchant = Merchant(**kwargs)

        self.invoice: Invoice = Invoice(**kwargs)

        self.qr: Qr = Qr(**kwargs)

        self.wallet: Wallet = Wallet(**kwargs)

    def get_api_token(self):
        return self.__api_token

# Async Monobank Open API


class AsyncMonobankOpenAPI:
    """
    Asynchronous version MonobankOpenAPI
    Source: https://api.monobank.ua/docs/
    """

    def __init__(self, api_token: Optional[str] = None):
        super().__init__(api_token=api_token, _async=True)


# Async Monobank Corporate Open API


class AsyncMonobankCorporateOpenAPI:
    """
    Asynchronous version MonobankCorporateOpenAPI
    Source: https://api.monobank.ua/docs/corporate.html
    """

    def __init__(self):
        pass

